/*
 * Copyright (C) 2022 Loongson Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * Authors:
 *    Sui Jingfeng <suijingfeng@loongson.cn>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_LSX
#include <lsxintrin.h>
#endif

#include "loongson_debug.h"
#include "etnaviv_resolve.h"

/* TODO: provide a none simd version */

Bool lsx_resolve_etnaviv_tile_4x4(uint32_t *src_bits,
                                  uint32_t *dst_bits,
                                  int src_stride,
                                  int dst_stride,
                                  int src_x,
                                  int src_y,
                                  int dest_x,
                                  int dest_y,
                                  int width,
                                  int height)
{
    TRACE_ENTER();

#ifdef HAVE_LSX
    int src_stride_tiled = src_stride * 4;
    int dst_stride_tiled = dst_stride * 4;
    int i, j;

    dst_bits += dst_stride * dest_y + dest_x;
    src_bits += src_stride * src_y + src_x;

    DEBUG_MSG("%s: src stride=%d, dst stride=%d, src addr: %p, dst addr: %p\n",
            __func__, src_stride, dst_stride, src_bits, dst_bits);

    for (i = 0; i < height; i += 4)
    {
        uint32_t *pSrc = src_bits;
        uint32_t *pDst0 = dst_bits;
        uint32_t *pDst1 = pDst0 + dst_stride;
        uint32_t *pDst2 = pDst1 + dst_stride;
        uint32_t *pDst3 = pDst2 + dst_stride;
        int offset = 0;

        for (j = 0; j < width; j += 4)
        {
            __m128i v0, v1, v2, v3;

            /* load one tile, tile is stored in continues row */
            v0 = __lsx_vld(pSrc, 0);
            v1 = __lsx_vld(pSrc, 16);
            v2 = __lsx_vld(pSrc, 32);
            v3 = __lsx_vld(pSrc, 48);

            __lsx_vstx(v0, pDst0, offset);
            __lsx_vstx(v1, pDst1, offset);
            __lsx_vstx(v2, pDst2, offset);
            __lsx_vstx(v3, pDst3, offset);

            offset += 16;
            pSrc += 16;
        }

        src_bits += src_stride_tiled;
        dst_bits += dst_stride_tiled;
    }
#endif

    TRACE_EXIT();

    return TRUE;
}
